from beepy import local_storage, trackable, utils, types
from beepy import listeners, attrs, children
from beepy import context, framework, style, tags, actions, router, import_hooks
from beepy.tags import Head, Body
from beepy.style import Style
from beepy.context import *
from beepy.context import __all__ as __context_all__
from beepy.framework import *
from beepy.framework import __version__, __all__ as __framework_all__

__all__ = [
    'local_storage', 'trackable', 'utils', 'types',
    'listeners', 'attrs', 'children',
    'context', 'framework', 'style', 'tags', 'actions', 'router', 'import_hooks',
    'Head', 'Body', 'Style',
    *__context_all__,
    *__framework_all__,
]



#!/usr/bin/env python3
import os

try:
    import websockets as _
except ImportError:
    print('Did you forget to install dev-requirements.txt?')
    exit(1)

import sys
import time
import asyncio
import http.server
import socketserver
from threading import Thread
from websockets.server import serve
from websockets.exceptions import ConnectionClosedOK
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler


class Manager:
    def __init__(self):
        self.websockets = []
        self.root_path = ''
        self.observer = None

    def set_ws(self, websocket):
        self.websockets.append(websocket)

    async def ws_send(self, message):
        for ws in self.websockets[:]:
            try:
                await ws.send(message)
            except ConnectionClosedOK:
                self.websockets.remove(ws)

        if not self.websockets:
            print('No clients connected! Please, restart your page to connect to the dev server')

    def set_root_path(self, path):
        self.root_path = path

    def set_observer(self, observer):
        self.observer = observer


m = Manager()


# File Watcher


class MonitorFolder(FileSystemEventHandler):
    def on_any_event(self, event):
        if event.is_directory or event.src_path.endswith(('~', '.tmp')) or event.event_type in ('opened', 'closed'):
            return

        path = event.src_path
        if path.startswith(m.root_path):
            path = path[len(m.root_path):]
        if path.startswith('.idea'):
            return

        asyncio.run(m.ws_send(path))


def start_watcher():
    root_path = sys.argv[1] if len(sys.argv) > 1 else os.getcwd()

    event_handler = MonitorFolder()
    observer = Observer()
    observer.schedule(event_handler, path=root_path, recursive=True)
    print(f'Monitoring started for {root_path}')
    observer.start()
    m.set_root_path(root_path)
    m.set_observer(observer)
    try:
        while (True):
            time.sleep(1)

    except KeyboardInterrupt:
        observer.stop()
        observer.join()


# WebSocket Server


async def echo(websocket):
    m.set_ws(websocket)

    async for message in websocket:
        print(f'{websocket=} {message=}')  # We really don't receive messages


async def main_ws():
    print('WebSockets started')
    async with serve(echo, 'localhost', 8998):
        await asyncio.Future()  # run forever


def start_websockets():
    asyncio.run(main_ws())


HTTP_PORT = 8888
Handler = http.server.SimpleHTTPRequestHandler


def start_http():
    with socketserver.TCPServer(('', HTTP_PORT), Handler) as httpd:
        print(f'Serving at port {HTTP_PORT}\nOpen server: http://localhost:{HTTP_PORT}')
        httpd.serve_forever()


Thread(target=start_watcher, daemon=True).start()
Thread(target=start_websockets, daemon=True).start()
Thread(target=start_http, daemon=True).start()
while True:
    time.sleep(60)



import re
from typing import Any, Optional, Type

import js

from beepy.framework import __CONFIG__, Tag, attr, state
from beepy.utils import log10_ceil, get_random_name, to_kebab_case, safe_issubclass
from beepy.tags import Head
from beepy.types import safe_html


def dict_of_properties_to_css(properties):
    for prop, value in properties.items():
        if not prop:
            yield value
            continue
        if not isinstance(prop, (tuple, set)):
            prop = (prop,)
        result = ''
        for _property in prop:
            _property = to_kebab_case(_property)
            result += '    ' + _property
            if isinstance(value, (list, tuple)):  # handle raw css
                result += ' ' + (_property + ' ').join(x for x in value if x)
            else:
                if value == '':
                    value = '""'
                result += f': {value};'
        yield result


def dict_of_parents_to_css(children, parent, braces):
    for child, value in children.items():
        if not child:
            yield f' {braces[0]} {child.strip()} {braces[1]}'
        if isinstance(child, (tuple, set)):
            child = ','.join(child)
        if '&' in child:
            child = re.sub('&', parent, child)
        else:
            child = parent + ' ' + child
        for inner in dict_to_css(value, child, braces):
            if child.startswith('@') and not parent:
                inner = f' {braces[0]} {inner.strip()} {braces[1]}'
            yield inner


def dict_to_css(selectors: dict, parent: str = '', braces=('{', '}')):
    """
    parses nested dict to css

    dict(
        div=dict(
            color='red',
            backgroundColor='blue',
            p={
                'font-size': '20px',
                '&.class': dict(
                    font_size='30px',
                ),
                '': '''
                    // some raw css
                '''
            }
        )
    )

    to

    div {
        color: red;
        background-color: blue;
    }
    div p {
        font-size: 20px;
                    // some raw css
    }
    div p.class {
        font-size: 30px;
    }


    """

    # TODO: implement all scss or create wheel of libsass, add micropip.import and refactor styles

    children = {}
    properties = {}

    for prop, value in selectors.items():
        if isinstance(value, dict):
            children[prop] = value
        else:
            properties[prop] = value

    if properties:
        yield parent + ' ' + braces[0]
        yield from dict_of_properties_to_css(properties)
        yield braces[1]

    yield from dict_of_parents_to_css(children, parent, braces)


def get_reference(tag: Tag):
    return f'[style-id="{tag.style_id}"]'


class Style(Tag, name='style', content_tag=None, raw_html=True, force_ref=True):
    __slots__ = ('styles', 'real_parent')
    __extra_attrs__ = ('style_id',)

    _global = {
        'styles_count': 1,
    }

    options = state(type=dict)
    real_parent: Optional[Tag]

    @classmethod
    def from_css(cls, file):
        # TODO: implement import from css/scss/pycss
        raise AttributeError

    def __init__(self, styles=None, options=None, get_vars=None, **styles_dict):
        super().__init__()
        if styles and not styles_dict:
            styles_dict = styles
        if options is None:
            options = {}
        self.styles = styles_dict
        self._content = ''
        self.real_parent = None
        self.options = {
            'global': False,
            'render_states': True,
            'render_children': True,
            'get_vars_callback': get_vars,
        } | options

    def __mount__(self, element, parent, index=None):
        self.real_parent = parent
        if __CONFIG__['style_head']:
            super().__mount__(Head.mount_element, Head)
        else:
            super().__mount__(element, parent, index)

    def __unmount__(self, element, parent, _unsafe=False):
        return super().__unmount__(element, parent, _unsafe=True)

    def mount(self):
        parent = self.real_parent

        if self.options['global']:  # TODO: add example
            self._content = '\n'.join(list(dict_to_css(self.styles, parent._tag_name_)))
            return

        if hasattr(parent, 'style_id'):  # support multiple style children
            style_id = parent.style_id
        else:
            self._global['styles_count'] += 1
            style_id = get_random_name(log10_ceil(self._global['styles_count']) + 1)
            attr(style_id, type=str)._link_ctx('style_id', parent)

        # TODO: is [style-id=] slow? If so, maybe use classes?
        self._content = '\n'.join(
            list(dict_to_css(self.styles, f'{parent._tag_name_}[style-id="{style_id}"]', braces=('{{', '}}')))
        )

    @safe_html.content
    def content(self):
        if self.options['global'] or not self._content:
            return self._content

        parent = self.real_parent

        params: dict[str, Any] = {}
        if self.options['render_states']:
            params.update(parent.__states__)
        if self.options['render_children']:
            params.update(parent.ref_children)
        if get_vars := self.options['get_vars_callback']:
            params.update(get_vars(self=parent, ref=get_reference, **params))

        # TODO: use native css '--var: {}' instead of re-render the whole content
        return self._content.strip().format(**params)

    @classmethod
    def import_file(cls, file_path):
        return js.beepy.addElement(
            js.document.head,
            'link',
            href=js.beepy.getPathWithCurrentPathAndOrigin(file_path),
            onload=cls.create_onload(),
            type='text/css',
            rel='stylesheet',
        )


def with_style(style_or_tag_cls: Optional[Style | Type[Tag]] = None):
    """
    @with_style
    class Button(Tag, name='button'):
        ...

    styled_button = with_style(Style(display='block', **kwargs))(button)
    """

    def wrapper(tag_cls: Type[Tag]):
        return type(tag_cls.__name__, (tag_cls,), {'style': style_or_tag_cls or Style()})

    if safe_issubclass(style_or_tag_cls, Tag):
        _tag_cls, style_or_tag_cls = style_or_tag_cls, None
        return wrapper(_tag_cls)

    return wrapper


__all__ = ['dict_to_css', 'Style', 'with_style']



from __future__ import annotations

import inspect
import traceback
from collections import defaultdict
from typing import Union, Callable, Type, Any, Iterable, Optional
from types import MethodType
from functools import wraps, partial, cache
from copy import deepcopy

import js

from beepy.attrs import attr, state, html_attr
from beepy.children import CustomWrapper, StringWrapper, ContentWrapper, ChildRef, TagRef, Children
from beepy.listeners import on
from beepy.types import safe_html, Renderer, Mounter, WebBase, AttrType, ContentType
from beepy.utils import (
    log, NONE_TYPE, _PY_TAG_ATTRIBUTE, __CONFIG__, _debugger, get_random_name, to_kebab_case, IN_BROWSER, to_js
)
from beepy.context import OVERWRITE, SUPER, CONTENT, _SPECIAL_CHILD_STRINGS, _MetaContext, Context


__version__ = '0.6.1'
__CONFIG__['version'] = __version__


if IN_BROWSER:
    js.Element.__str__ = lambda self: f'<{self.tagName.lower()}/>'
_current__lifecycle_method: dict[str, dict[int, 'Tag']] = {}


def _lifecycle_method(*, hash_function=hash):
    def _wrapper(fn):
        name = fn.__name__
        attr_name = f'_wrapper_{name}_calling'
        _cache = _current__lifecycle_method[attr_name] = {}

        @wraps(fn)
        def lifecycle_method(original_func):
            @wraps(original_func)
            def original_method_wrapper(self, *args, **kwargs):
                # prevent calling super() calls extra code twice
                _hash = hash_function(self)
                not_in_super_call = _hash not in _cache

                if not_in_super_call:
                    _cache[_hash] = self
                    result = fn(self, args, kwargs, _original_func=original_func)
                    del _cache[_hash]
                else:
                    result = original_func(self, *args, **kwargs)

                return result
            return original_method_wrapper
        return lifecycle_method
    return _wrapper


_TAG_INITIALIZED = False


class _MetaTag(_MetaContext):
    _tag_classes: list[Type[Tag]] = []
    __clean_class_attribute_names = ('_content_tag', '_static_children_tag')

    def __new__(mcs, _name: str, bases: tuple, namespace: dict, **kwargs):
        # TODO: create decorator for functions like this:
        # @tag
        # def test_tag(self: Tag, counter: int = 0):
        #     @self.on
        #     def click(event):
        #         self.counter += 1
        #
        #     @self.renderer
        #     def content():
        #         return f'Counter: {self.counter}'

        namespace = namespace.copy()
        namespace.setdefault('__slots__', ())
        if '_content' in namespace:
            namespace['_static_content'] = namespace.pop('_content')

        # used for checking inheritance: attributes, methods, etc.
        # for example: extending classes Tag and WithRouter must produce correct state 'router'
        base_cls: Union[Type[Tag], type] = type.__new__(mcs, _name, bases, {})

        initialized = _TAG_INITIALIZED  # if class Tag is already defined

        is_root = kwargs.get('_root')
        if is_root:
            tag_name = ''
        else:
            tag_name = kwargs.get('name')
        namespace['__ROOT__'] = is_root

        if tag_name or (initialized and not hasattr(base_cls, '_tag_name_')):
            namespace['_tag_name_'] = to_kebab_case(tag_name or _name)

        if 'raw_html' in kwargs:
            namespace['_raw_html'] = kwargs['raw_html']

        if 'force_ref' in kwargs:
            namespace['_force_ref'] = kwargs['force_ref']

        if 'mount' in kwargs:
            namespace['mount_element'] = kwargs['mount']

        if 'content_tag' in kwargs:
            content_tag = kwargs['content_tag']
            if isinstance(content_tag, str):
                content_tag = empty_tag(content_tag)()
            elif not isinstance(content_tag, (Tag, NONE_TYPE)):
                raise TypeError('content_tag argument must be either str, Tag or None')
            namespace['_content_tag'] = content_tag

        if 'children_tag' in kwargs:
            children_tag = kwargs['children_tag']
            if isinstance(children_tag, str):
                children_tag = empty_tag(children_tag)()
            elif not isinstance(children_tag, Tag):
                raise TypeError('children_tag argument must be either str or Tag')
            namespace['_static_children_tag'] = children_tag

        ref_children = []
        to_remove_children = []
        static_onchange_handlers = []

        super_children_index = -1
        children_arg = namespace.get('children', [])
        if isinstance(children_arg, property):
            children_arg = []

        if children_arg:
            namespace.pop('children')
            children_arg = list(children_arg)
            if OVERWRITE not in children_arg:
                if SUPER not in children_arg:
                    children_arg.insert(0, SUPER)
                super_children_index = children_arg.index(SUPER)
        elif children_arg not in (None, False, []):
            namespace['_static_children'] = namespace.pop('children')
            children_arg = []

        if initialized:
            for attribute_name, child in tuple(mcs._clean_namespace(namespace)):
                if callable(child) and hasattr(child, '_attrs_static_'):
                    _states_with_static_handler = defaultdict(list)
                    for trigger, _states in child._attrs_static_.items():
                        for _state in _states:
                            _states_with_static_handler[trigger].append(_state)
                            _state.handlers[trigger].remove(child)
                    static_onchange_handlers.append((child, _states_with_static_handler))
                    continue

                if isinstance(child, (Tag, Children, ChildRef)):
                    if child in children_arg:
                        ref_children.append(child)
                        # TODO: make possible inherit without replacement?
                        if (old_child := getattr(base_cls, attribute_name, None)) and isinstance(old_child, ChildRef):
                            to_remove_children.append(old_child)
                        if isinstance(child, (Tag, Children)):
                            children_arg[children_arg.index(child)] = namespace[attribute_name] = child.as_child(None)
    
                    if isinstance(child, Tag) and child._force_ref:  # TODO: add support '_force_ref' for Children too
                        namespace[attribute_name] = child.as_child(None)
                        children_arg.append(namespace[attribute_name])

            for _index, child in enumerate(children_arg):
                if isinstance(child, str) and child not in _SPECIAL_CHILD_STRINGS:
                    children_arg[_index] = StringWrapper(child)

                if isinstance(child, Tag) and child not in ref_children:
                    # TODO: replace 5 in get_random_name(5) with some log
                    children_arg[_index] = namespace[f'_{get_random_name(5)}_'] = child.as_child(None)

        cls: Union[Type[Tag], type] = super().__new__(mcs, _name, bases, namespace, **kwargs)

        if not hasattr(cls, '_raw_html'):
            cls._raw_html = False

        if initialized:
            cls._static_children = cls._static_children.copy()
            if not hasattr(cls, '_content_tag'):
                cls._content_tag = empty_div()
        else:
            cls._static_children = []

        if not hasattr(cls, '_static_children_tag'):
            cls._static_children_tag = None

        for child in to_remove_children:
            cls._static_children.remove(child)

        if getattr(cls, '_tag_name_', None) or not isinstance(getattr(cls, 'children', None), property):
            if children_arg:
                if CONTENT in children_arg and CONTENT in cls._static_children:
                    cls._static_children.remove(CONTENT)
                children_arg[super_children_index: super_children_index + 1] = cls._static_children
                cls._static_children = children_arg
        else:
            cls._static_children = [CONTENT]

        if initialized:
            cls._static_listeners = deepcopy(cls._static_listeners)
            cls._static_onchange_handlers = cls._static_onchange_handlers.copy() + static_onchange_handlers
        else:
            cls._static_listeners = defaultdict(list)
            cls._static_onchange_handlers = []

        cls._tags = []

        mcs._tag_classes.append(cls)

        if '__mount__' in namespace:
            cls.__mount__ = mcs.__mount(cls.__mount__)

        if '__render__' in namespace:
            cls.__render__ = mcs.__render(cls.__render__)

        if '__init__' in namespace:
            cls.__init__ = mcs.__init(cls.__init__)

        if '__unmount__' in namespace:
            cls.__unmount__ = mcs.__unmount(cls.__unmount__)

        if initialized and 'mount' in kwargs:
            cls._root_parent = None
            setattr(cls.mount_element, _PY_TAG_ATTRIBUTE, cls())

        if cls.__ROOT__:
            cls.__root_declared__()
        else:
            cls.__class_declared__()

        return cls

    @_lifecycle_method()
    def __mount(self: Tag, args, kwargs, _original_func):
        self.__class__._tags.append(self)

        self._mount_finished_ = False

        result = _original_func(self, *args, **kwargs)

        self._mount_attrs()

        for child in self.children:
            if isinstance(child, CustomWrapper):
                child.__mount__(self.mount_element, self)
            elif isinstance(child, Mounter):
                child.__mount__(self._children_element, self)
            else:
                log.warn(f'Cannot mount: {child}')
        if self._children_tag:
            self.mount_element.insertChild(self._children_element)

        if self.content_child is None:
            _debugger(self)
        else:
            self.content_child._mount_children()

        for name, attribute in self.__states__.items():
            if callable(attribute) and not isinstance(attribute, MethodType):
                setattr(self, name, MethodType(attribute, self.parent))

        self._post_mount_attrs()

        for event, listeners in self._listeners.items():
            for listener in listeners:
                self._event_listeners[event].append(
                    listener._make_listener(event, self)
                )

        for onchange_handler, _states_with_static_handler in self._static_onchange_handlers:
            for trigger, _states in _states_with_static_handler.items():
                for _state in _states:
                    # TODO: can we save order of triggers' call?
                    _state.handlers[trigger].append(MethodType(onchange_handler, self))

        self.mount()

        self._mount_finished_ = True

        return result

    @_lifecycle_method()
    def __render(self: Tag, args, kwargs, _original_func):
        if not self._mount_finished_ or (
            not self.mount_element.parentElement and self._root_parent != self  # dismounted
        ):
            return  # Prevent render before mount finished; Could be useful for setting intervals inside mount method

        # TODO: maybe function 'render' could return some content, appended to args?
        self.render()

        _original_func(self, *args, **kwargs)

        self.post_render()

    @_lifecycle_method(hash_function=id)
    def __init(self: Tag, args, kwargs, _original_func):
        called_from_clone = kwargs.pop('__parent__', None)
        if hasattr(getattr(self, 'mount_element', None), _PY_TAG_ATTRIBUTE):
            return

        self._children = self._static_children.copy()
        self._content = self._static_content

        children_argument: Union[Callable, ContentType] = kwargs.get('children') or args
        if children_argument and (not isinstance(children_argument, Iterable) or isinstance(children_argument, str)):
            children_argument = (children_argument,)
        children_argument = list(children_argument)

        if children_argument:
            is_child_arg_string = False
            is_child_arg_function = False
            is_child_arg_tag = False
            try:
                content_empty = not bool(self.content())
            except Exception as e:
                content_empty = False
            if content_empty:
                is_child_arg_string = True
                is_child_arg_function = True
                is_child_arg_tag = True
                for index, child_arg in enumerate(children_argument):
                    if not isinstance(child_arg, str):
                        is_child_arg_string = False
                    if (isinstance(child_arg, type) and issubclass(child_arg, Tag)) or not callable(child_arg):
                        is_child_arg_function = False
                    if not isinstance(child_arg, Tag):
                        if isinstance(child_arg, str) and len(children_argument) > 1:
                            children_argument[index] = empty_span(child_arg)
                        else:
                            is_child_arg_tag = False

            if is_child_arg_string and children_argument:
                self._content = children_argument
            elif is_child_arg_function and children_argument and len(children_argument) == 1:
                self._content = MethodType(children_argument[0], self)
            elif is_child_arg_tag:
                self._children = self._children.copy() + [
                    child.clone(self).as_child(self) for child in children_argument
                ]
            else:
                self._children = self._children.copy() + [*children_argument]

        for key, value in kwargs.items():
            if not isinstance(value, (Tag, Children)) or key in self.attrs:
                continue
            child = value.as_child(self, exists_ok=True)
            child.__set_name__(self, key)
            setattr(type(self), key, child)
            self._children.append(child)

        self._dependents = []
        self._shadow_root = None
        self._parent_ = None
        self.mount_parent = None
        self._mount_finished_ = False

        if not hasattr(self, 'mount_element'):
            self.mount_element = js.document.createElement(self._tag_name_)
        if getattr(self.mount_element, _PY_TAG_ATTRIBUTE, None):
            raise ValueError(f'Coping or using as child is not allowed for "{self._tag_name_}"')
        else:
            setattr(self.mount_element, _PY_TAG_ATTRIBUTE, self)
        if self._static_children_tag:
            self._children_tag = self._static_children_tag.clone(self)
            self._children_element = self._children_tag.mount_element
            setattr(self._children_element, _PY_TAG_ATTRIBUTE, self)
        else:
            self._children_tag = None
            self._children_element = self.mount_element

        self._listeners = deepcopy(self._static_listeners)

        self._event_listeners = defaultdict(list)
        self._handlers = defaultdict(list)
        self._ref = None

        _original_func(self, *args, **kwargs)

    @_lifecycle_method()
    def __unmount(self: Tag, args, kwargs, _original_func):
        if self in self.__class__._tags:  # But why?
            self.__class__._tags.remove(self)

        self.unmount()

        for child in self.children:
            if isinstance(child, CustomWrapper):
                child.__unmount__(self.mount_element, self)
            elif isinstance(child, Mounter):
                child.__unmount__(self._children_element, self)
        if self._children_tag:
            self.mount_element.safeRemoveChild(self._children_element)

        result = _original_func(self, *args, **kwargs)

        if IN_BROWSER:
            for event, event_listeners in self._event_listeners.items():
                for event_listener in event_listeners:
                    on._remove_listener(event, self, event_listener)

        self.post_unmount()

        return result


class Tag(WebBase, Context, metaclass=_MetaTag, _root=True):
    # TODO: add docstrings

    __slots__ = (
        '_content', '_dependents', '_shadow_root', '_ref',
        '_listeners', '_event_listeners',
        'mount_parent', '_parent_', '_mount_finished_',
        'mount_element', '_children', '_children_element', '_children_tag',
        '_handlers',
    )

    _root_parent: Tag = None  # see function mount in the bottom

    _static_content: ContentType = ''

    _content: ContentType
    _dependents: list[Renderer]
    _shadow_root: js.HTMLElement
    _ref: Optional[TagRef]
    _force_ref: bool = False

    _tag_name_: str
    _tags: list[Tag]
    _static_listeners: defaultdict[str, list[on]]
    _listeners: defaultdict[str, list[on]]
    _event_listeners: defaultdict[str, list[Callable[[js.Event], None]]]
    mount_element: js.HTMLElement
    mount_parent: js.HTMLElement
    parent: Optional[Tag]
    _static_children: list[ContentType]
    _children: list[Mounter]
    _children_element: js.HTMLElement
    _static_children_tag: Tag
    _children_tag: Tag
    _mount_finished_: bool
    _handlers: defaultdict[str, list[Callable[[Tag, js.Event, str, Any], None]]]
    _static_onchange_handlers: list[tuple[Callable[[Tag, Any], Any], dict[str, list[state]]]]

    @classmethod
    def __root_declared__(cls):
        """ This method is called, when root Tag is defined """

    @classmethod
    def __class_declared__(cls):
        """ This method is called, when common Tag is defined """

    def __init__(self, *args, **kwargs: AttrType):
        # DO NOT DELETE; This method must be wrapped by _MetaTag.__init
        kwargs.setdefault('_load_children', False)
        super().__init__(*args, **kwargs)

    def __new__(cls, *args, **kwargs):
        if hasattr(getattr(cls, 'mount_element', None), _PY_TAG_ATTRIBUTE):
            self = getattr(cls.mount_element, _PY_TAG_ATTRIBUTE)
        else:
            self = super().__new__(cls, *args, **kwargs)

        return self

    def __repr__(self):
        return f'{type(self).__name__}(<{self._tag_name_}/>)'

    def __hash__(self):
        # TODO: make force immutable this attributes
        return hash((self._tag_name_, self._id_))

    def __getitem__(self, key):
        return getattr(self, key)

    def as_child(self, parent: Optional[Tag], exists_ok=False, inline_def=False):
        if self._ref:
            if exists_ok:
                self.__set_ref__(parent, self._ref)
                return self._ref
            else:
                raise TypeError(f'Tag {self._tag_name_} already is child')
        ref = TagRef(self, inline_def=inline_def)
        self.__set_ref__(parent, ref)
        return ref

    def __notify__(self, attr_name: str, attribute: attr, value: AttrType):
        super().__notify__(attr_name, attribute, value)
        self.__render__()

    def __set_ref__(self, parent: Optional[Tag], ref: TagRef):
        self._ref = ref
        if ref.inline_def:
            setattr(type(parent), ref.name, self)

    def render(self):
        """empty method for easy override with code for run before render"""

    def __render__(self, attrs: Optional[dict[str, AttrType]] = None):
        self._current_render.append(self)

        if attrs is None:
            attrs = {}

        for name, value in {**self.__attrs__, **attrs}.items():
            # TODO: optimize this - set only changed attributes
            if getattr(self.mount_element.attributes, name, None) and value is None:
                self.mount_element.removeAttribute(name)
            elif value is not None:
                self.mount_element.setAttribute(name, value)

        content_index = self.get_content_index()
        if content_index is not None:
            self.children[content_index].__render__()

        for index, child in enumerate(self.children):
            # TODO: optimize this - re-render the only children, that need this
            if isinstance(child, ChildRef):
                child.__render__(self)
            elif isinstance(child, CustomWrapper):
                if content_index is None or index != content_index:
                    child.__render__()
            elif isinstance(child, Renderer):
                child.__render__()
            else:  # string ?
                self._children_element.append(self._render(child))

        if self._current_render[-1] is self:
            self._current_render.pop()

    def post_render(self):
        """empty method for easy override with code for run after render"""

    def content(self) -> ContentType:
        return self._content() if callable(self._content) else self._content

    def _mount_attrs(self):
        for attribute in self.attrs.values():
            attribute.__mount_tag__(self)

    def _post_mount_attrs(self):
        for attribute in self.attrs.values():
            attribute.__post_mount_tag__(self)

    def pre_mount(self):
        """empty method for easy override with code for run before mount"""

    def init(self, *args, _load_children=True, **kwargs):
        super().init(*args, **kwargs)

        if _load_children:
            self.children = self.children.copy()

        for name, value in self.ref_children.items():
            if name not in kwargs:
                continue

            if isinstance(value, Children):
                getattr(self, name)[:] = kwargs[name]

    def __mount__(self, element, parent: Tag, index=None):
        self.parent = parent
        self.mount_parent = element
        self.pre_mount()

        args, kwargs = self.args_kwargs
        kwargs = self._attrs_defaults | kwargs
        self.init(*args, **kwargs)

        self.mount_parent.insertChild(self.mount_element, index)

    def mount(self):
        """empty method for easy override with code for run after mount"""

    @property
    def parent_defined(self):
        return self._parent_ is not None

    @property
    def parent(self):
        if self._parent_ is None:
            try:
                raise ValueError
            except ValueError:
                frame = inspect.currentframe().f_back
                log.warn(traceback.format_exc(), inspect.getsourcefile(frame), frame.f_lineno, to_js(frame.f_locals))
        return self._parent_

    @parent.setter
    def parent(self, v):
        self._parent_ = v

    def unmount(self):
        """empty method for easy override with code for run before unmount"""

    def __unmount__(self, element, parent, _unsafe=False):
        if not _unsafe and self.mount_parent is not None and self.mount_parent is not element:
            log.warn(
                'Something went wrong!\n'
                f'Real parent: {self.mount_parent} {self.parent}. Passed parent: {element} {parent} '
                'If you override __mount__, you also should override __unmount__ too.'
            )
            log.warn(''.join(traceback.format_stack()[:-1]))

        (self.mount_parent or element).safeRemoveChild(self.mount_element)

    def post_unmount(self):
        """empty method for easy override with code for run after unmount"""

    @property
    def own_children(self) -> list[Mounter]:
        return [child for child in self.children if not isinstance(child, (ChildRef, CustomWrapper))]

    @property
    def ref_children(self) -> dict[str, Mounter]:
        return {child.name: child.__get__(self) for child in self.children if isinstance(child, ChildRef)}

    @property
    def content_child(self) -> ContentWrapper:
        for child in self.children:
            if isinstance(child, ContentWrapper) and child.content.__func__.__name__ == 'content':
                return child

    def get_content_index(self):
        for index, child in enumerate(self.children):
            if isinstance(child, ContentWrapper) and child.content.__func__.__name__ == 'content':
                return index

    @property
    def children(self) -> list[Union[Mounter, Renderer, ChildRef, ContentWrapper]]:
        return self._children

    @children.setter
    def children(self, children):
        self._children = self._handle_children(children)

    def _handle_children(self, children):
        result = []

        for child in children:
            result.append(child)

            if isinstance(child, ContentWrapper):
                child = child.content.__func__
                if child.__name__ == 'content':
                    child = CONTENT

            if child == CONTENT:
                child = self.content

            if isinstance(child, attr):
                _attr_name = child.name

                def child(s):
                    return getattr(s, _attr_name)

            if isinstance(child, Tag) and child in self._children and child._ref is not None:
                # using Tag as descriptor/just child; this allows save reference from parent to new copy of child
                child._ref._update_child(self, self._children.index(child))
            elif isinstance(child, ChildRef) and child in self._children:
                child._update_child(self, self._children.index(child))
            # TODO: make ContentWrapper descriptor, here call .clone()
            #  make possible to use ContentWrapper separately from function 'content'
            elif callable(child):
                if not isinstance(child, MethodType):
                    child = MethodType(child, self)
                result[-1] = ContentWrapper(child, self._content_tag, self._current_render)

        return result

    def clone(self, parent=None) -> Tag:
        clone = super().clone(parent=parent)
        clone._children = self.children
        clone._listeners = deepcopy(self._listeners)
        clone._handlers = deepcopy(self._handlers)
        return clone

    def on(self, method: Union[Tag, str]):
        if isinstance(method, str) and method.startswith(':'):  # TODO: maybe it could be useful in `class on()`?
            action = method[1:]

            def wrapper(handler, action_name=None):
                if action_name is None:
                    action_name = handler.__name__

                self._handlers[action_name].append(handler)
                return handler

            if action:
                return partial(wrapper, action_name=action)

            return wrapper

        def wrapper(callback):
            event_listener = on(method)(callback, get_parent=True)
            event_name = event_listener.name or callback.__name__
            event_listener.__set_name__(self, event_name, set_static_listeners=False)
            self._listeners[event_name] = self._listeners[event_name].copy() + [event_listener]
            return event_listener.__get__(self)

        if not isinstance(method, str):
            return wrapper(method)

        return wrapper

    @property
    def _current_render(self):
        try:
            return _MetaTag._current_render[self._root_parent]
        except KeyError as e:
            _debugger(e)
            raise ValueError('It looks like element is not mounted correctly, please see the docs')


_TAG_INITIALIZED = True


@cache
def empty_tag(name):
    return _MetaTag(name, (Tag, ), {}, name=name, content_tag=None)


def inline_tag(name, content_tag=None, **attributes):
    return _MetaTag(name, (Tag,), attributes, name=name, content_tag=content_tag)


empty_div = empty_tag('div')
empty_span = empty_tag('span')


def mount(element: Tag, root_element: str, clear=False):
    root = js.document.querySelector(root_element)
    if root is None:
        raise NameError('Mount point not found')

    js.beepy.stopLoading()
    if clear or js.beepy.DEV__hot_reload:
        root.innerHTML = ''
    js.beepy.startLoading(mountPoint=root)

    parent = inline_tag(root.tagName.lower(), _root_parent=state(type=Tag, move_on=True))()
    parent._root_parent = parent
    parent._attrs_defaults['_root_parent'] = parent
    parent.__class__._root_parent.initial_value = parent
    parent.mount_element = root
    element.link_parent_attrs(parent)

    setattr(root, _PY_TAG_ATTRIBUTE, parent)
    _MetaTag._top_mount(element, root, parent)
    _MetaTag._top_render(element)

    if not js.document.title:
        js.document.title = 'BeePy'
        log.warn(f'Document title is not set, use default title: {js.document.title}')


__all__ = [
    '__version__', '__CONFIG__', '_debugger', 'attr', 'state', 'html_attr', '_MetaTag', 'Tag', 'on', 'empty_tag',
    'mount', 'safe_html',
]



import sys

from importlib.abc import MetaPathFinder
from importlib.util import spec_from_file_location

import js
from beepy.utils import IN_BROWSER, __CONFIG__, _debugger


BEEPY_ROOT_PACKAGE = '__beepy_root__'
requirements = __CONFIG__['requirements']
MODULES_NOT_EXISTING_ON_SERVER = [
    BEEPY_ROOT_PACKAGE,
    *(requirements() if callable(requirements) else requirements),
    '_hashlib',
    '_strptime',
    'unicodedata',
    'pprint',
]
# some modules must be ignored to prevent load it from local server, when importing modules like micropip or datetime
# TODO: move this hook higher to support importing `beepy/__init__.py` natively


class ServerFinder(MetaPathFinder):
    def find_spec(self, fullname, path, target=None):
        if path and any(p.startswith('/lib') for p in path):
            return

        if fullname in MODULES_NOT_EXISTING_ON_SERVER:
            return

        is_beepy_module = fullname.startswith('beepy.')
        current_path = js.beepy.__CURRENT_LOADING_FILE__
        if is_beepy_module:
            js.beepy.__CURRENT_LOADING_FILE__ = __CONFIG__['path']

        try:
            err = js.beepy._loadLocalModule(fullname, checkPathExists=True)
        except Exception as e:
            err = e

        if err or is_beepy_module:
            js.beepy.__CURRENT_LOADING_FILE__ = current_path

        if err:
            _debugger(err)
            MODULES_NOT_EXISTING_ON_SERVER.append(fullname)
            return

        return spec_from_file_location(fullname)


if IN_BROWSER:
    sys.meta_path.insert(0, ServerFinder())
    sys.path.append(BEEPY_ROOT_PACKAGE)



from __future__ import annotations

from abc import abstractmethod
from typing import Optional, Callable, Union, Type, TypeVar, Generic, Iterable

import js

import beepy
from beepy.types import Tag, Renderer, WebBase, Children, ContentType
from beepy.context import Context
from beepy.utils import log, _PY_TAG_ATTRIBUTE


class CustomWrapper(WebBase):
    pass


class StringWrapper(CustomWrapper):
    __slots__ = ('content',)

    def __init__(self, string):
        self.content = string

    def __mount__(self, element, parent: Tag, index=None):
        element.insertChild(self._render(self.content).format(self=parent), index)

    def __repr__(self):
        return f'String({self.content})'


class ContentWrapper(CustomWrapper):
    __slots__ = ('content', 'tag', 'mount_element', '_current_render', 'parent', 'mount_parent', 'children')

    SHADOW_ROOTS = (
        'article', 'aside', 'blockquote', 'body', 'div', 'footer', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'header',
        'main', 'nav', 'p', 'section', 'span',
    )

    content: Callable[[], Union[ContentType, Tag]]
    tag: Optional[Tag]
    mount_element: Optional[js.HTMLElement]
    _current_render: list[Renderer]
    parent: Optional[Tag]
    mount_parent: Optional[js.HTMLElement]
    children: Optional[list[Tag]]

    def __init__(self, content, tag, _current_render):
        self.content = content
        self.tag = tag
        if tag:
            self.mount_element = tag.mount_element
        else:
            self.mount_element = None

        self._current_render = _current_render
        self.parent = None
        self.mount_parent = None
        self.children = None

    def __mount__(self, element, parent: Tag, index=None):
        self.parent = parent
        self.mount_parent = element
        parent_name = element.tagName.lower()
        if self.tag:
            self.mount_element = self.tag.clone(parent).mount_element
            self.mount_parent.insertChild(self.mount_element, index)
        elif ('-' in parent_name or parent_name in self.SHADOW_ROOTS) and False:
            # TODO: add some param to use shadow root directly
            # (or) TODO[fix]: with shadow root children cannot be rendered
            # (or) TODO: at least warn developer, or raise error, if not fixed
            if self.parent._shadow_root:
                self.mount_element = self.parent._shadow_root
            else:
                self.mount_element = self.mount_parent.attachShadow(mode='closed')
            self.parent._shadow_root = self.mount_element
        else:
            self.mount_element = js.document.createDocumentFragment()
            self.mount_parent.insertChild(self.mount_element, index)
        setattr(self.mount_element, _PY_TAG_ATTRIBUTE, self)

    def _mount_children(self):
        content = self.content()

        if isinstance(content, beepy.framework.Tag):
            content = (content,)
        elif isinstance(content, Iterable) and not isinstance(content, str) and content:
            content = list(content)
            for _child in content[:]:
                if not isinstance(_child, beepy.framework.Tag):
                    content = None
                    break
        else:
            content = None

        if content:
            for child in content:
                child.__mount__(self.mount_element, self.parent)
            self.children = content

    def __unmount__(self, element, parent):
        if self.mount_parent is not element:
            return

        if self.tag:
            element.safeRemoveChild(self.mount_element)
        # TODO: handle shadowRoot and documentFragment?

    def __render__(self):
        for renderer in self._current_render:
            if self.parent and renderer not in self.parent._dependents:
                self.parent._dependents.append(renderer)

        self._current_render.append(self)

        if self.children:
            for child in self.children:
                child.__render__()

            if self._current_render[-1] is self:
                self._current_render.pop()
            return

        result = self._render(self.content())
        if not isinstance(result, str):
            raise TypeError(f'Function {self.content} cannot return {result}!')

        if self.tag or (self.parent and self.parent._shadow_root):
            self.mount_element.innerHTML = result
        else:  # fragment can't be re-rendered
            self.mount_element.innerHTML = result
            current_html = self.mount_parent.innerHTML
            current_html_escaped = self._render(current_html)
            if result and result not in (current_html, current_html_escaped):
                if current_html and not (self.parent and self.parent._raw_html):
                    log.warn(
                        f'This html `{current_html}` will be replaces with this: `{result}`.\n'
                        'Maybe you must use beepy.Tag instead of beepy.tags.div, '
                        'or you used incorrect html tags like <br/> instead of <br>',
                    )
                self.mount_parent.innerHTML = result

        if self._current_render[-1] is self:
            self._current_render.pop()

    def __repr__(self):
        return f'<{self.parent}.{self.content.__name__}()>'


C = TypeVar('C')


class ChildRef(WebBase, Generic[C]):
    __slots__ = ('name', 'child', 'inline_def', '_cache')

    name: Optional[str]
    child: C
    _cache: dict[Context, C]

    def __init__(self, child: C, inline_def=False):
        self.name = None
        self.child = child
        self.inline_def = inline_def
        self._cache = {}

    def __repr__(self):
        return f'{type(self).__name__}(Tag.{self.name} = {self.child})'

    def __get__(self, instance: Optional[Context], owner: Optional[Type[Context]] = None) -> Union[ChildRef, C]:
        if instance is None:
            return self

        if (result := self._cache.get(instance)) is not None:
            return result

        self._cache[instance] = self.child
        return self.child

    def __set__(self, instance: Context, value: C):
        self._cache[instance] = value

    def __delete__(self, instance: Context):
        del self._cache[instance]

    def __set_name__(self, owner: Type[Tag], name: str):
        self.name = name

    @abstractmethod
    def _update_child(self, parent: Tag, index: int):
        pass

    def __render__(self, parent: Tag):
        self.__get__(parent).__render__()

    def __mount__(self, element, parent: Tag, index=None):
        self.__get__(parent).__mount__(element, parent, index)

    def __unmount__(self, element, parent: Tag):
        self.__get__(parent).__unmount__(element, parent)


class TagRef(ChildRef[Tag]):
    __slots__ = ()

    child: Tag

    def _update_child(self, parent, index):
        clone = self.__get__(parent).clone(parent)
        clone.__set_ref__(parent, self)
        self.__set__(parent, clone)


class ChildrenRef(ChildRef):
    __slots__ = ()

    child: Children

    def _update_child(self, parent, index):
        copy = self.__get__(parent).copy()
        copy.__set_parent__(parent, index, self)
        self.__set__(parent, copy)

    def __set__(self, instance, value):
        if isinstance(value, Children):
            super().__set__(instance, value)
        else:
            current_value = self.__get__(instance)
            current_value[:] = value


from beepy.framework import Tag


__all__ = ['CustomWrapper', 'StringWrapper', 'ContentWrapper', 'TagRef', 'ChildrenRef']



from __future__ import annotations

import inspect
from typing import Optional, Any, Callable
from types import MethodType
from functools import partial, wraps
from copy import deepcopy

import js

from beepy.types import Tag
from beepy.utils import _PY_TAG_ATTRIBUTE, create_proxy, to_js

GLOBAL_EVENTS_LIST = ('keyup', 'keypress', 'keydown')
_key_codes = {
    'esc': (27,),
    'tab': (9,),
    'enter': (13,),
    'space': (32,),
    'up': (38,),
    'left': (37,),
    'right': (39,),
    'down': (40,),
    'delete': (8, 46),
}


# TODO: click.(right|middle) ; click.left ; etc.


def key_code_check(key_name, event):
    return event.keyCode in _key_codes[key_name]


RAW_JS_CHECKS = ('prevent', 'stop', 'stop_all')


class on:
    __slots__ = ('name', 'callback', 'get_parent', 'modifiers', 'checks', 'js_checks')

    name: Optional[str]
    callback: Callable[[Tag, ...], Any]
    get_parent: bool
    modifiers: list[str]
    checks: list[Callable[[js.Event], bool]]
    js_checks: list[str]

    def __init__(self, method):
        self.get_parent = False
        self.modifiers = []
        self.checks = []
        self.js_checks = []

        if isinstance(method, str):
            if '.' in method:
                method, *self.modifiers = method.split('.')
                for modifier in self.modifiers:
                    if modifier in _key_codes:
                        # TODO: check for visibility?
                        self.checks.append(partial(key_code_check, modifier))
                    elif modifier in RAW_JS_CHECKS:
                        self.js_checks.append(modifier)
                    else:
                        raise ValueError(f'Unknown event modifier ".{modifier}"!')

            self.name = method
            return

        self.name = None
        self(method)

    def __call__(self, method, get_parent=None):
        self.callback = method
        self.get_parent = get_parent
        return self

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        return self.callback

    def _prepare_call(self, tag, event):
        for check in self.checks:
            if not check(event):
                return

        if self.get_parent:
            tag = tag.parent

        if isinstance(self.callback, MethodType):
            return self.callback, tag
        else:
            return MethodType(self.callback, tag), tag

    async def _a_call(self, tag, event):
        if (prepare := self._prepare_call(tag, event)) is None:
            return

        fn, tag = prepare

        data = await fn(event)
        self._after_call(tag, event)
        return data

    def _call(self, tag, event):
        if (prepare := self._prepare_call(tag, event)) is None:
            return

        fn, tag = prepare
        data = fn(event)
        self._after_call(tag, event)
        return data

    def _after_call(self, tag, event):
        for dependent in tag._dependents:
            # TODO: move to other place
            dependent.__render__()

        getattr(event.currentTarget, _PY_TAG_ATTRIBUTE, tag).__render__()

    def _make_listener(self, event_name: str, tag: Tag):
        if inspect.iscoroutinefunction(self.callback):
            @wraps(self.callback)
            async def method(event):
                return await self._a_call(tag, event)
        else:
            @wraps(self.callback)
            def method(event):
                return self._call(tag, event)
        is_global = event_name in GLOBAL_EVENTS_LIST  # TODO: check this || what about set global by attr?
        return js.beepy.addAsyncListener(
            js.document if is_global else tag.mount_element, event_name, create_proxy(method), to_js(self.js_checks)
        )

    @classmethod
    def _remove_listener(cls, event_name: str, tag: Tag, event_listener: Callable):
        is_global = event_name in GLOBAL_EVENTS_LIST

        (js.document if is_global else tag.mount_element).removeEventListener(event_name, event_listener)

    def __set__(self, instance, value):
        raise AttributeError(f'Cannot set on_{self.name} event handler')

    def __delete__(self, instance):
        pass

    def __set_name__(self, owner, name, *, set_static_listeners=True):
        if self.name is None:
            self.name = name

        if set_static_listeners:
            owner._static_listeners = deepcopy(owner._static_listeners)
            owner._static_listeners[self.name].append(self)

    def __repr__(self):
        return f'on_{self.name}({self.callback})'


__all__ = ['on', '_key_codes', 'RAW_JS_CHECKS']



from http.client import HTTPException
import inspect
import dataclasses
import json
import math
import re
import sys
import string
import random
# TODO: add some param to use shadow root directlyimport inspect
import traceback
import asyncio
from functools import wraps
from typing import Any
from datetime import datetime
from importlib import import_module

import js
import micropip
from pyodide import http as pyodide_http
import beepy


# TODO: make it a module (could be a quite complicated due to loading beepy modules as /<file>.py requests)


try:
    from pyodide.ffi import IN_BROWSER, create_once_callable, create_proxy, to_js as pyodide_to_js
except ImportError:
    from pyodide import IN_BROWSER, create_once_callable, create_proxy, to_js as pyodide_to_js
    js.console.log('FutureWarning: You need to upgrade pyodide version to 0.21.0 or higher')


class UpgradedJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        if isinstance(o, datetime):
            return o.strftime(__CONFIG__['default_datetime_format'])
        return super().default(o)


def to_js(obj, dict_converter=js.Object.fromEntries, **kwargs):
    return pyodide_to_js(obj, dict_converter=dict_converter, **kwargs)


log = js.console
NONE_TYPE = type(None)
_PY_TAG_ATTRIBUTE = '__PYTHON_TAG__'
__CONFIG__ = {
    'debug': False,
    'style_head': True,
    'api_url': '/',
    'default_datetime_format': '%Y-%m-%dT%H:%M:%S.%f%Z',
    'inputs_auto_id': True,
    'html_replace_whitespaces': True,
    'modules': [],
    'requirements': [],
}

if 1 == 1:
    __CONFIG__['debug'] = True


def merge_configs():
    __CONFIG__.update(js.beepy.__CONFIG__.to_py())
    js.beepy.__CONFIG__ = to_js(__CONFIG__)


if IN_BROWSER:  # TODO: check support for non-browser runs
    merge_configs()


def _debugger(error=None):
    if isinstance(error, Exception):
        log.warn(traceback.format_exc())
    else:
        log.warn(''.join(traceback.format_stack()[:-1]))
        log.warn(error)
    frame = inspect.currentframe().f_back
    js._locals = to_js(frame.f_locals)
    js._locals._frame = frame
    js._DEBUGGER(error)


def log10_ceil(num):
    return math.ceil(math.log10(num or 1)) or 1


def wraps_with_name(name):
    def wrapper(func):
        func.__name__ = func.__qualname__ = name
        return func
    return wrapper


def escape_html(s, quote=False, whitespace=False):
    """ Replace special characters "&", "<" and ">" to HTML-safe sequences.
        If the optional flag quote is True (default), the quotation mark characters (" and ') are also translated.
        If the optional flag whitespace is True, new line (\n) and tab (\t) characters are also translated.
    """
    #       Must be done first!
    s = s.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;')
    if quote:
        s = s.replace('"', '&quot;').replace('\'', '&#39;')
    if whitespace:
        s = s.replace('\n', '<br>').replace('\t', '&emsp;').replace('  ', ' &nbsp;')
    return s


_w = string.ascii_letters + string.digits + '_'


def get_random_name(length=6):
    return ''.join(
        random.choice(_w)
        for _ in range(length)
    )


def to_kebab_case(name: str, *, replacer='-'):
    """
    parsing name of tag to html-compatible or name of property to css-compatible
    >>> class __beepy__(Tag): ...  # <beepy></beepy>
    >>> class myTagName(Tag): ...  # <my-tag-name/>
    >>> Style(font_size='20px')  # font-size: 20px
    >>> Style(backgroundColor='red')  # background-color: red
    """
    return re.sub(
        r'(?P<upper>[A-Z])',
        lambda m: replacer + m.group('upper').lower(),
        re.sub('[_ ]', replacer, name)
    ).strip(replacer)


def _need_update(old_func, old_name='', *, version):
    old_name = old_name or ('beepy.utils.' + to_kebab_case(old_func.name, replacer='_'))

    @wraps(old_func)
    def _wrapper(*args, **kwargs):
        log.warn(
            f'FutureWarning: function {old_name} is slightly wrapped for old versions of pyodide. '
            f'You need to upgrade pyodide version to {version} or higher'
        )
        new_args = []
        for arg in args:
            if callable(arg):
                arg = create_once_callable(arg)
            new_args.append(arg)
        return old_func(*new_args, **kwargs)

    return _wrapper


try:
    from pyodide.ffi.wrappers import (
        set_timeout, clear_timeout, set_interval, clear_interval, add_event_listener, remove_event_listener,
    )
except ImportError:
    set_timeout = _need_update(js.setTimeout, version='0.21.0')
    clear_timeout = _need_update(js.clearTimeout, version='0.21.0')
    set_interval = _need_update(js.setInterval, version='0.21.0')
    clear_interval = _need_update(js.clearInterval, version='0.21.0')
    add_event_listener = _need_update(js.addEventListener, version='0.21.0')
    remove_event_listener = _need_update(js.removeEventListener, version='0.21.0')


class const_attribute(property):
    def __set__(self, instance, value):
        if self.__get__(instance) is None:
            super().__set__(instance, value)
        else:
            raise AttributeError


def safe_issubclass(type_or_Any, class_or_tuple_to_check):
    try:
        return issubclass(type_or_Any, class_or_tuple_to_check)
    except TypeError:
        return False


class AnyOfType:
    __slots__ = ('type',)

    def __init__(self, type):
        self.type = type

    def __eq__(self, other):
        return isinstance(other, self.type)

    def __repr__(self):
        return f'AnyOfType<{self.type}>'

    def __hash__(self):
        return hash(self.type)


async def request(url, method='GET', body=None, headers=None, **opts):
    if body is not None:
        body = json.dumps(body, cls=UpgradedJSONEncoder)

    if headers is None:
        headers = {}

    headers.update({
        'mode': 'no-cors',
        'Content-Type': 'application/json',
        'Access-Control-Allow-Headers': '*',
        'Access-Control-Allow-Origin': '*',
    })

    response = await pyodide_http.pyfetch(
        __CONFIG__['api_url'] + url, method=method, body=body, headers=headers, **opts
    )

    if int(response.status) >= 400:
        raise HTTPException(response.status)

    if method == 'GET' or opts.get('to_json'):
        response = await response.json()
    else:
        response = await response.string()

    return response


if not IN_BROWSER:
    import requests

    async def request(url, method='GET', body=None, headers=None, **opts):
        if body is not None:
            body = json.dumps(body, cls=UpgradedJSONEncoder)

        if headers is None:
            headers = {}

        # TODO: check opts argument compatibility

        response = requests.request(
            method, __CONFIG__['api_url'] + url, data=body, headers=headers,
        )

        if int(response.status_code) >= 400:
            raise HTTPException(response.status_code)

        if method == 'GET' or opts.get('to_json'):
            response = response.json()
        else:
            response = response.text

        return response


def ensure_sync(to_await, callback=lambda x: x):
    if inspect.iscoroutine(to_await):
        task = asyncio.get_event_loop().run_until_complete(to_await)
        task.add_done_callback(callback)
        return task

    callback(to_await)
    return to_await


def force_sync(function):
    # TODO: maybe do it with class will be better?

    callbacks = []

    @wraps(function)
    def wrapper(*args, _done_callback_=None, **kwargs):
        current_callbacks = [(cb() if dynamic else cb) for cb, dynamic in callbacks]
        current_callbacks.append(_done_callback_)

        def _callback(_res_):
            try:
                r = _res_.result()
            except Exception as e:
                _debugger(e)
                return

            for cb in current_callbacks:
                if cb is None:
                    continue
                try:
                    cb(*args, **kwargs, _res_=r)
                except Exception as e:
                    _debugger(e)
        return ensure_sync(function(*args, **kwargs), _callback)

    wrapper.run_after = wrapper.add_callback = lambda fn: callbacks.append((fn, False)) or fn
    wrapper.add_dynamic_callback = lambda fn: callbacks.append((fn, True)) or fn

    return wrapper


def force_sync__wait_load(function):
    wrapper = force_sync(function)
    wrapper.add_dynamic_callback(beepy.context.Context.create_onload)
    return wrapper


force_sync.wait_load = force_sync__wait_load


delay = js.delay


@force_sync
async def sleep(s):  # TODO: check if this actually works or not
    return await js.delay(s * 1000)


class Interval:
    __slots__ = ('_id',)

    def __init__(self, function, args=None, kwargs=None, period=0):
        @wraps(function)
        def _callback():
            return function(*(args or ()), **(kwargs or {}))

        self._id = set_interval(_callback, period * 1000)

    def clear(self):
        return clear_interval(self._id)


class Locker:
    __slots__ = ('name', 'locked')

    def __init__(self, name='Lock'):
        self.name = name
        self.locked = False

    def __enter__(self):
        self.locked = True

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.locked = False

    def __bool__(self):
        return self.locked

    def __str__(self):
        return f'Locker<{self.name}>({self.locked})'


async def reload_requirements():
    get_requirements = __CONFIG__['requirements']
    if not callable(get_requirements):  # static requirements, must be already loaded
        return

    await asyncio.gather(*[
        micropip.install(requirement)
        for requirement in get_requirements()
    ])


def push_url(url, unused='', **url_state):
    js.history.pushState(to_js({'href': url.href, **url_state}), unused, url.href)


def replace_url(url, unused='', **url_state):
    js.history.replaceState(to_js({'href': url.href, **url_state}), unused, url.href)


def lazy_import(module_path):
    if not (
        (module := sys.modules.get(module_path))
        and (spec := getattr(module, "__spec__", None))
        and getattr(spec, "_initializing", False) is False
    ):
        module = import_module(module_path)

    return module


def import_string(dotted_path):
    module_path, class_name = dotted_path.rsplit('.', 1)
    return getattr(lazy_import(module_path), class_name)


def lazy_import_cls(cls):
    if isinstance(cls, str):
        return import_string(cls)
    return cls


class _BeePyGlobals(dict):
    def __getitem__(self, key):
        result = super().__getitem__(key)
        for handler in __beepy_global_handlers__:
            result = handler(self, key, result)
        return result


def _default_global_handlers(locals_dict, key, result):
    if key.startswith('_p__'):
        return result(locals_dict)
    return result


__beepy_global_handlers__ = [_default_global_handlers]


__all__ = [
    'IN_BROWSER', 'create_once_callable', 'create_proxy', 'to_js',
    'log', '_PY_TAG_ATTRIBUTE', 'NONE_TYPE', '__CONFIG__', '_debugger',
    'log10_ceil', 'wraps_with_name', 'get_random_name', 'to_kebab_case',
    'set_timeout', 'clear_timeout', 'set_interval', 'clear_interval', 'add_event_listener', 'remove_event_listener',
    'const_attribute', 'ensure_sync', 'force_sync', 'delay', 'sleep',
    'lazy_import', 'import_string', 'lazy_import_cls', 'safe_issubclass', 'reload_requirements',
]
